"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const http = require("http");
const express = require("express");
const request = require("request");
const socketio = require("socket.io");
const uuid = require("uuid/v4");
const path = require("path");
const compression = require("compression");
const middleware = require("./middleware");
const jsonLines_1 = require("./jsonLines");
const liveSharing_1 = require("./liveSharing");
var OutdatedPackage;
(function (OutdatedPackage) {
    function fromTableRow(row) {
        return {
            name: row[0],
            current: row[1],
            wanted: row[2],
            latest: row[3],
            type: row[4],
        };
    }
    OutdatedPackage.fromTableRow = fromTableRow;
})(OutdatedPackage = exports.OutdatedPackage || (exports.OutdatedPackage = {}));
var ServerCommand;
(function (ServerCommand) {
    ServerCommand["addPackage"] = "addPackage";
    ServerCommand["removePackage"] = "removePackage";
    ServerCommand["publishPackage"] = "publishPackage";
    ServerCommand["reloadClients"] = "reloadClients";
    ServerCommand["updatePackageInfo"] = "updatePackageInfo";
    ServerCommand["setIsDocumentDirty"] = "setIsDocumentDirty";
    ServerCommand["upgradeDependency"] = "upgradeDependency";
    ServerCommand["upgradeFramerLibrary"] = "upgradeFramerLibrary";
    ServerCommand["updateCanvasData"] = "updateCanvasData";
    ServerCommand["startLiveSharing"] = "startLiveSharing";
    ServerCommand["stopLiveSharing"] = "stopLiveSharing";
    ServerCommand["joinRoom"] = "joinRoom";
    ServerCommand["updatePreview"] = "updatePreview";
    ServerCommand["metrics"] = "metrics";
    ServerCommand["buildStandaloneExport"] = "buildStandaloneExport";
})(ServerCommand = exports.ServerCommand || (exports.ServerCommand = {}));
class Server {
    constructor(project) {
        this._state = {
            url: undefined,
            filename: undefined,
            running: false,
            packages: [],
            outdated: [],
            packageErrors: [],
            serverErrors: [],
            tasks: {},
            isDocumentDirty: false,
            liveSharing: liveSharing_1.LiveSharing.initialState(),
        };
        this.previewData = null;
        this.rawCanvasData = null;
        const app = express();
        this.http = http.createServer(app);
        this.io = socketio(this.http);
        this.project = project;
        this.commandHandler = project.commandHandler;
        this.liveSharing = new liveSharing_1.LiveSharing("http://framer.live");
        this.liveSharing.on("statusUpdate", () => {
            this.setState({ liveSharing: this.liveSharing.state });
        });
        this.liveSharing.on("error", error => {
            Server.logError(`Live Preview Error: ${error}`);
        });
        app.use(middleware.nocache);
        app.use(middleware.logging);
        app.use(middleware.cors);
        app.use(compression());
        const socketPath = path.join(__dirname, "..", "node_modules/socket.io-client/dist");
        app.use("/_socket", express.static(socketPath));
        // must make sure design images are fully cachable
        app.use("/design/images", express.static(path.join(project.path, "design", "images"), {
            setHeaders: function (res, path) {
                res.setHeader("Cache-Control", "public, max-age=36000");
                res.setHeader("Pragma", "");
                res.setHeader("Surrogate-Control", "");
            },
        }));
        // normal images are under user control, should invalidate after a while
        app.use("/images", express.static(path.join(project.path, "images"), {
            setHeaders: function (res, path) {
                res.setHeader("Cache-Control", "public, must-revalidate, max-age=36000");
                res.setHeader("Pragma", "");
                res.setHeader("Surrogate-Control", "");
            },
        }));
        // redirect / and /index.html to preview/preview.html
        if (project.staticPath) {
            // This works because the static middleware will call next when not found
            app.get("/", function (req, res) {
                if (req.path === "/" || req.path === "/index.html") {
                    res.redirect("/preview/");
                }
            });
            // Needed to make the route above work
            app.use("/preview", express.static(path.join(project.staticPath, "Vekter"), { index: "preview.html" }));
            app.use("/_app/resources", express.static(project.staticPath));
        }
        app.use("/", express.static(project.path));
        // Use for debugging, example usage
        app.get("/_loader", (req, res) => {
            res.send(`

            <script src="/_socket/socket.io.js"></script>
            <script>var socket = io(); socket.on("script", function(script) {

                const require = (name) => {
                    throw Error("Component loader: Can't require '" + name + "' likely an external in webpack")
                }

                const fn = new Function("exports", "require", script)
                const exportedModules = {}

                fn(exportedModules, require)

                console.log("package",  exportedModules)

            });
            </script>
          `);
        });
        app.use("/_proxy", function (req, res) {
            req.pipe(request(req.url.slice(1))).pipe(res);
        });
        Server.instance = this;
        this.setup();
    }
    get state() {
        return this._state;
    }
    static logError(error) {
        const instance = Server.instance;
        const queue = Server.errorsQueue;
        if (!instance) {
            queue.push(error);
            return;
        }
        // Sending the queue first
        if (queue.length) {
            queue.forEach((error) => {
                instance.sendError(error);
            });
            queue.length = 0;
        }
        instance.sendError(error);
    }
    setup() {
        this.io.on("connect", socket => {
            Object.keys(ServerCommand).forEach((command) => {
                socket.on(command, info => {
                    if (!this.handleServerCommand(socket, command, info)) {
                        this.commandHandler(command, info);
                    }
                });
            });
            socket.on("resetServerErrors", () => this.setState({ serverErrors: [] }));
            // Send the state over the just connected socket
            socket.emit("state", this.state);
            // Send the script at load (for dev), but only to the socket connecting!
            this.project.sendScript(socket);
            if (this.state.serverErrors.length) {
                socket.emit("serverErrors", { errors: this.state.serverErrors });
            }
        });
    }
    // Returns true if handled by server
    handleServerCommand(socket, command, info) {
        switch (command) {
            // Hijack commands that don’t really involve the project
            case "startLiveSharing": {
                this.liveSharing.start(this.http.address().port);
                return true;
            }
            case "stopLiveSharing": {
                this.liveSharing.stop();
                return true;
            }
            case "joinRoom": {
                socket.join(info.room);
                if (info.room == "preview") {
                    if (this.rawCanvasData) {
                        socket.emit("updateCanvasData", this.rawCanvasData);
                    }
                    if (this.previewData) {
                        socket.emit("updatePreview", this.previewData);
                    }
                }
                return true;
            }
            case "updatePreview": {
                this.previewData = info.previewData || null;
                this.io.to("preview").emit("updatePreview", info.previewData);
                return true;
            }
            case "updateCanvasData": {
                this.rawCanvasData = info.canvasData || null;
                this.io.to("preview").emit("updateCanvasData", info.canvasData);
                return false; // Because we still want the project to handle this command too!
            }
            case "metrics": {
                this.io.emit("metrics", { type: info.type });
                return true;
            }
            case "buildStandaloneExport": {
                if (!this.project)
                    return true;
                this.project
                    .buildStandalone(info.path, info.previewData)
                    .then(() => {
                    socket.emit("buildStandaloneExportDone", "ok");
                })
                    .catch(error => {
                    console.log(error);
                    socket.emit("buildStandaloneExportDone", error && error.message ? error.message : error);
                });
                return true;
            }
            default: {
                return false;
            }
        }
    }
    start(port = 3000) {
        return __awaiter(this, void 0, void 0, function* () {
            let tryPort = typeof port === "number" ? port : port[0];
            const maxPort = typeof port === "number" ? port : port[1];
            // When setting a host (which we want, to prevent leaking info to the network),
            // we also need to set “exclusive”, else the server will end up pooling with other
            // servers.
            const listenOptions = { host: "127.0.0.1", exclusive: true };
            const onError = error => {
                if (error.code === "EADDRINUSE" && tryPort < maxPort) {
                    this.http.listen(Object.assign({}, listenOptions, { port: ++tryPort })); // Retry
                }
                else {
                    Promise.reject(error);
                }
            };
            const onListening = () => {
                const { address, port } = this.http.address();
                this.http.removeListener("error", onError);
                this.setState({
                    url: `http://${address}:${port}`,
                    filename: path.basename(this.project.path),
                    running: true,
                });
                this.liveSharing.enable();
                Promise.resolve();
            };
            this.http.on("error", onError);
            this.http.once("listening", onListening);
            this.http.listen(Object.assign({}, listenOptions, { port: tryPort }));
        });
    }
    registerTask(name, task, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const taskId = uuid();
            this.addTaskToState(taskId, name, context);
            try {
                yield task;
            }
            finally {
                this.removeTaskFromState(taskId);
            }
        });
    }
    runAsTask(name, task, context) {
        return __awaiter(this, void 0, void 0, function* () {
            // Although registerTask is async, we intentionally don't wait for it here
            this.registerTask(name, task, context);
            return task;
        });
    }
    addTaskToState(id, name, context) {
        const tasks = Object.assign({}, this.state.tasks);
        tasks[id] = context ? { name, context } : { name };
        this.setState({ tasks });
    }
    removeTaskFromState(id) {
        const tasks = Object.assign({}, this.state.tasks);
        delete tasks[id];
        this.setState({ tasks });
    }
    setState(state) {
        Object.assign(this.state, state);
        // Log to stdout, used by the app (at least for the first state)
        console.log(JSON.stringify(this.state));
        // Send the state over all connected sockets
        this.io.emit("state", this.state);
    }
    sendError(error) {
        try {
            const yarnErrors = jsonLines_1.JSONLines.parse(error);
            let errors = yarnErrors
                // Keep only error type
                .filter(item => item.type === "error")
                // Filter out errors summary 'Found N errors.'
                .filter(item => !item.data.includes("Found "));
            if (!yarnErrors.length && typeof error === "string" && error.length) {
                // Filter out a Yarn bug we often run into
                if (!error.includes("/node_modules/yarn/lib/cli.js")) {
                    errors = [{ data: error, type: "error" }];
                }
            }
            this.setState({ serverErrors: this.state.serverErrors.concat(errors) });
            this.io.emit("serverErrors", { errors });
        }
        catch (err) {
            console.log("Error happened while sending error!");
        }
    }
}
Server.errorsQueue = [];
exports.Server = Server;
